from . import helpers, pwpeg


class AstNode(object):
    def __repr__(self):
        return "{0}".format(self.__class__.__name__)

    def accept(self, visitor, *args, **kwargs):
        methodname = "visit_" + self.__class__.__name__
        return getattr(visitor, methodname)(self, *args, **kwargs)


class AstProduction(AstNode):
    def __init__(self, code=None, repetition=None):
        self.label = ""
        self.code = code
        self.repetition = None
        self.matching = None

    def set_label(self, label):
        self.label = label
        return self

    def set_repetition(self, repetition):
        self.repetition = repetition
        return self

    def set_matching(self, matching):
        self.matching = matching
        return self

    def __repr__(self):
        label = ""
        repetition = self.repetition or ""
        if self.label:
            label = self.label + ":"
        return "{0}{1}{2}".format(label, self.code, repetition)


class AstRuleCall(AstProduction):
    def __init__(self, decl=None, repetition=None):
        self.decl = decl
        self.label = decl.label
        self.repetition = None
        self.code= ""


class AstLookAhead(AstProduction):
    def __init__(self, production, symbol):
        super(AstLookAhead, self).__init__()
        self.production = production
        self.symbol = symbol


class AstPredicate(AstProduction):
    def __init__(self, code):
        super(AstPredicate, self).__init__()
        self.code = code

    def __repr__(self):
        return "{{{0}}}".format(self.code)



class AstProductionGroup(AstProduction):
    def __init__(self, rules):
        super(AstProductionGroup, self).__init__()
        self.rules = rules
        self.action = None

    def set_action(self, action):
        self.action = action
        return self

    def __repr__(self):
        rules = " ".join([repr(r) for r in self.rules])

        action = ""

        if self.action:
            action = " ({0})".format(repr(self.action))

        return "{0}{1}".format(rules, action)


class AstProductionChoices(AstProductionGroup):
    def __repr__(self):
        rules = " | ".join([repr(r) for r in self.rules])

        action = ""

        if self.action:
            action = " ({0})".format(repr(self.action))

        return "{0}{1}".format(rules, action)



class AstRuleDeclaration(AstNode):
    def __init__(self, name):
        self.name = name
        self.args = None
        self.productions = None
        self.skip = None
        self.label = name

    def set_productions(self, productions):
        self.productions = productions
        return self

    def set_args(self, args):
        self.args = args
        return self

    def set_skip(self, skip):
        self.skip = skip
        return self

    def __repr__(self):
        return "{0}{1}".format(self.name, self.args)


class AstCode(AstNode):
    def __init__(self, code):
        self.code = code

    def __repr__(self, ctx=dict(), indent=0):
        code = ""
        if self.code:
            code = self.code[:10]
            if len(self.code) > 10:
                code += "..."
        return code


class AstFile(AstNode):

    def __init__(self, code_start, rules, code_end):
        self.code_start = code_start or ""
        self.rules = rules
        self.code_end = code_end or ""

    def __repr__(self):
        return "{0}{1}{2}".format(self.code_start, self.rules, self.code_end)



"""
    @author Christophe Eymard <christophe@ravelsoft.com>
"""

from re import _pattern_type
import inspect
import sys

if sys.version_info >= (3, 0):
    # Python 3 removed entirely the unicode type, so to have
    # the input isinstance() still work, we just remap it to str (which
    # is unicode anyways)
    unicode = str
u = unicode


class SyntaxError(Exception):
    """ The way the input is parsed is by trial and error.

        When SyntaxError is caught by the parser, it goes up the caller
        chain to get to the first Choice() it finds to keep looking
        for a match.
    """
    def __init__(self, error, input, suberrors=[]):
        self.suberrors = suberrors
        self.line = input.line
        self.column = input.column
        self.pos = input.pos
        super(SyntaxError, self).__init__(error + "({0}:{1})".format(self.line, self.column))


    def fullmessage(self):
        return unicode(self) + "\n" + "\n".join([ "\n".join(["   " + line for line in e.fullmessage().split("\n")]) for e in self.suberrors])


class IgnoreResult(object):
    """ This class is used by the parsing rules to determine wether to add the
        result of some parsing to the general result.

        Typically, the And and Not rules return this as well as and advancement
        of 0 when they match.
    """


class Results(list):
    """
    """

    def __init__(self, name=""):
        self.name = name
        self.dict = {}

    def __repr__(self):
        return u("{0}:{1}").format(self.name, super(Results, self).__repr__())

    def add(self, name, value):
        self.dict[name] = len(self)
        self.append(value)

    def get(self, name, default=None):
        return self[self.dict[name]]


class Input(object):
    def __init__(self, input):
        self.input = input
        self.pos = 0
        self.line = 1
        self.column = 1

    def rewind(self, n):
        self.pos -= n

    def rewind_to(self, pos):
        self.rewind(self.pos - pos)

    def has_next(self):
        return self.pos < len(self.input)

    def current(self):
        return self.input[self.pos] if self.has_next() else None


class TextInput(Input):
    def startswith(self, s):
        if self.input.startswith(s, self.pos):
            self.advance(s)
            return s
        return None

    def match(self, re):
        m = re.match(self.input, self.pos)
        if not m:
            return None
        matched = m.group()
        self.advance(matched)
        return matched

    def advance(self, s):
        if not s: return

        l = len(s)
        self.line += s.count("\n")
        self.column = l - s.rfind("\n") - 1
        self.pos += l

    def rewind(self, n):
        """
        """

        self.line -= self.input.count("\n", self.pos - n, self.pos)
        self.pos -= n
        self.column = self.pos - self.input.rfind("\n", 0, self.pos)



class TokenInput(Input):
    # FIXME Create token input.

    def __init__(self, tokens, input):
        pass

    def match(self, reg): raise Exception("Token Inputs can't match regexps")
    def startswith(self, txt): raise Exception("Token Inputs can't match strings")


class Rule(object):
    """ A Grammar rule.
    """

    @staticmethod
    def getrule(obj):
        """ Get the rule object corresponding to a given type.
        """

        if isinstance(obj, bytes):
            return StringRule(obj.decode('utf-8'))

        if isinstance(obj, unicode):
            return StringRule(obj)

        if isinstance(obj, _pattern_type):
            return RegexpRule(obj)

        if callable(obj):
            return Predicate(obj)

        return obj


    def __init__(self, *args):
        self.action = None
        self.name = ""
        self.productions = None

        if len(args) > 0:
            self.set_productions(*args)


    def set_productions(self, *args):
        """
        """

        if len(args) == 0:
            raise Exception("Can not have empty rules")

        if len(args) > 1 or not isinstance(args[0], Rule):
            self.productions = [Rule.getrule(r) for r in args]
        else:
            self.productions = args

        self.name = "_"
        self.post_subrule_name(", ".join([s.name for s in self.productions]))
        return self

    def post_subrule_name(self, productions_names):
        self.name = productions_names

    def set_skip(self, skip):
        self.skip = Rule.getrule(skip)
        return self

    def set_name(self, name):
        self.name = name
        return self


    def parse(self, input, currentresults=None, skip=None):
        """ Execute the rules
        """
        results = Results(self.name)

        if "skip" in self.__dict__: skip = self.skip

        if not self.productions:
            raise Exception("There are no productions defined for " + self.name)

        pos_save = input.pos

        for r in self.productions:
            subrule_result = None

            if skip:
                try:
                    skip.parse(input, Results())
                except SyntaxError as e:
                    # Nothing to skip.
                    pass

            try:
                r.parse(input, results, skip)
            except SyntaxError as e:
                input.rewind_to(pos_save)
                raise SyntaxError(u("In {0} ").format(self.name), input, [e])

        if self.action:
            currentresults.append(self.action(*results))
            return

        if len(results) == 1:
            currentresults.append(results[0])
        else:
            currentresults.append(results)


    def __repr__(self):
        return self.name


    def set_action(self, fn):
        """ Add an action function to the rules.

            This is to be used when the processor function is not directly given to
            the rule in its constructor.
        """
        if not fn:
            return self

        if inspect.getargspec(fn)[2]:
            raise Exception("Actions can't take kwargs")

        self.action = fn
        return self


class StringRule(Rule):
    def __init__(self, string):
        self.string = unicode(string)
        self.name = "\"" + self.string + "\""

    def parse(self, input, currentresults=None, skip=None):
        if input.startswith(self.string):
            currentresults.append(self.string)
        else:
            raise SyntaxError(u("Expected {0}, but found \"{1}\"").format(self.name, input.current()), input)


class RegexpRule(Rule):
    def __init__(self, regexp):
        self.regexp = regexp
        self.name = "/" + self.regexp.pattern + "/"

    def parse(self, input, currentresults=None, skip=None):
        match = input.match(self.regexp)
        if match is not None:
            currentresults.append(match)
        else:
            raise SyntaxError(u("Expected {0}, but found \"{1}\"").format(self.name, input.current()), input)


class Predicate(Rule):
    def __init__(self, fn):
        self.fn = fn
        self.name = u("Predicate {0}").format(fn.__name__)

    def parse(self, input, currentresults=[], skip=None):
        if self.fn(*currentresults) is False:
            # None actually is a valid result.
            raise SyntaxError(u("{0} was not satisfied").format(self.name), input)


class FunctionRule(Rule):
    """ A rule that takes arguments.

        Beware, set_skip and set_action should NOT be set directly on this rule.
    """

    class InstanciatedRule(Rule):
        def __init__(self, fn, name, args):
            self.fn = fn
            self.name = name
            self.args = args
            self.action = None
            self.rule = None
            # self.skip = None

        def parse(self, input, currentresults, skip):
            if not self.rule:
                self.rule = self.fn(*self.args).set_name("*" + self.name)

                if hasattr(self, "skip") and not hasattr(self.rule, "skip"):
                    self.rule.set_skip(self.skip)

            results = Results(self.name)
            self.rule.parse(input, results, skip)
            if self.action:
                currentresults.append(self.action(*results))
                return
            if len(results) == 1:
                currentresults.append(results[0])
            else:
                currentresults.append(results)


    def __init__(self, fn=None):
        if fn:
            self.set_fn(fn)
            self.name = fn.__name__
        else:
            self.fn = None
            self.name = "Param rule <no function>"
        self.action = None

    def set_fn(self, fn):
        args, varargs, keywords, defaults = inspect.getargspec(fn)
        if keywords:
            raise Exception("Function rules functions can only take simple args")

        self.fn = fn
        self.name = fn.__name__
        return self

    def instanciate(self, *args):
        arg_names = u("({0})").format(", ".join([repr(a) for a in args]))
        # r = self.fn(*args).set_name(self.name + arg_names)
        r = FunctionRule.InstanciatedRule(self.fn, self.name + arg_names, args)

        if self.action:
            r.set_action(self.action)

        if "skip" in self.__dict__:
            r.set_skip(self.skip)

        return r

    def parse(self, input, currentresults=None, skip=None):
        # Works if the rule doesn't need any arguments
        rule = self.instanciate()
        rule.parse(input, currentresults, skip)



class Repetition(Rule):
    """ A set of rules that can be repeated, like {m,n} in regular expressions.
    """


    def __init__(self, _from, _to, *args):
        self._from = _from
        self._to = _to
        self.rule = Rule(*args)
        super(Repetition, self).__init__()
        self.post_subrule_name(self.rule.name)


    def parse(self, input, currentresults=None, skip=None):
        results = Results(self.name)

        times = 0
        _from, _to = self._from, self._to

        save_pos = input.pos
        last_error = []

        while input.has_next() and (_to == -1 or times < _to):
            try:
                # Get the resultss.
                self.rule.parse(input, results, skip)
                times += 1
            except SyntaxError as e:
                last_error.append(e.suberrors[0])
                break


        if _from != -1 and times < _from:
            input.rewind_to(save_pos)
            raise SyntaxError(u("{1} needs to be repeated at least {0} times").format(_from, self.name), input, last_error)

        if self.action:
            # Actions in repetitions are in the form of lists.
            currentresults.append(self.action(results))
        else:
            currentresults.append(results)

    def post_subrule_name(self, sn):
        self.name = sn + u("<{0}, {1}>").format(self._from, self._to)



class OneOrMore(Repetition):
    """ A Repetition that acts like the + in regular expressions.
    """

    def __init__(self, *args, **kwargs):
        super(OneOrMore, self).__init__(1, -1, *args, **kwargs)

    def post_subrule_name(self, subn):
        self.name = u("[{0}]+").format(subn)


class ZeroOrMore(Repetition):
    """ A Repetition that acts like the * in regular expressions.
    """

    def __init__(self, *args, **kwargs):
        super(ZeroOrMore, self).__init__(0, -1, *args, **kwargs)

    def post_subrule_name(self, subn):
        self.name = u("[{0}]*").format(subn)


class Exactly(Repetition):
    """ A Repetition that wants to match excatly `times` elements.
    """

    def __init__(self, times, *args, **kwargs):
        super(Exactly, self).__imit__(times, times, *args, **kwargs)

    def post_subrule_name(self, subn):
        self.name = u("[{0}]<{1}>").format(subn, self._from)


class Optional(Repetition):
    """ The equivalent of the ? in regular expressions.
    """

    def __init__(self, *args, **kwargs):
        super(Optional, self).__init__(0, 1, *args, **kwargs)


    def parse(self, input, currentresults=None, skip=None):
        results = Results()
        super(Optional, self).parse(input, results, skip)

        if len(results[0]) == 0:
            currentresults.append(None)
        else:
            currentresults.append(results[0][0])

    def post_subrule_name(self, sn):
        self.name = "[" + sn + "]?"


class Not(Rule):
    """ Look ahead in the input. If no syntax error is received, then
        raise a SyntaxError.

        This rule is used to check that a rule does not apply on the input
        following the current position.

        It does *not* advance the parser position.
    """

    def parse(self, input, currentresults=None, skip=None):
        save_pos = input.pos
        results = Results()
        try:
            # Results are ignored.
            super(Not, self).parse(input, results, skip)
        except SyntaxError as e:
            # Couldn't match the next rule, which is what we want, so
            # we just restore the parser's position so it will continue
            # parsing as normal.
            return
        input.rewind_to(save_pos)
        raise SyntaxError(u("In <{0}> Matched \"{1}\"").format(self.name, results), input)

    def post_subrule_name(self, productions):
        self.name = "Not " + productions



class And(Rule):
    """ Look ahead in the input. If its rules can be applied, then continue
        parsing from the same place we were before executing this rule.

        This rule is used to check that the following input contains a rule we
        want.

        It does *not* advance the parser position.
    """


    def parse(self, input, currentresults=None, skip=None):
        results = Results() # We're going to ignore any results.
        save_pos = input.pos
        # Try to parse our rules
        super(And, self).parse(input, results, skip)

        # If there was no error, we don't advance, which is what we want.
        # A syntax error is raised otherwise in our super() parse.
        input.rewind_to(save_pos)

    def post_subrule_name(self, sn):
        self.name = u("Look-Ahead {0}").format(sn)



class Either(Rule):
    """ Try parsing with several rules and return the result of the first
        one that works.

        The rules are given to the constructor as its arguments.
    """

    def parse(self, input, currentresults=None, skip=None):
        all_errors = []
        results = Results()

        for rule in self.productions:
            try:
                rule.parse(input, results, skip)
                res = results[0]

                if self.action:
                    currentresults.append(self.action(res))
                else:
                    currentresults.append(res)
                return
            except SyntaxError as e:
                all_errors.append(e)
                # We continue since the SyntaxError just means that we didn't match and
                # must try the next choice.
                continue

        raise SyntaxError(u("In [{0}], none of the provided choices matched").format(self.name), input, all_errors)

    def post_subrule_name(self, subn):
        self.name = u("either({0})").format(subn)


class Any(Rule):
    def __init__(self):
        self.name = "Any"

    def parse(self, input, currentresults=None, skip=None):
        if skip is None and "skip" in self.__dict__: skip = self.skip

        save_pos = input.pos

        if skip:
            try:
                skip.parse(input, Results())
            except SyntaxError as e:
                # Nothing to skip.
                pass

        if not input.has_next():
            input.rewind_to(save_pos)
            raise SyntaxError("Want anything, but no more input", input)

        any = input.current()
        input.advance(any)
        currentresults.append(any)

class MemoRule(Rule):
    """ A rule that memorizes itself for future uses.
    """

    def __init__(self, rule):
        self.memorized = None
        self.rule = rule
        self.name = "Memorizing({0})".format(rule.name)

    def parse(self, input, currentresults=None, skip=None):
        """
        """
        if not self.memorized:
            # act = self.__dict__.get("action", None)
            # if act: del self.__dict__["action"]
            self.rule.parse(input, currentresults, skip)

            res = currentresults[len(currentresults) - 1]
            if not isinstance(res, list) and not isinstance(res, tuple):
                self.memorized = Rule(res)
            else:
                self.memorized = Rule(*res)
        else:
            # The rule is now memorized, and we can execute it.
            self.memorized.parse(input, currentresults, skip)


class Parser(object):
    """ A parser that parses a input input.

        It is given a top-level rule with which it will start the parsing,
        as well as a skip rule which will be checked agains before executing
        any rule, useful to remove white spaces and comments.
    """

    def __init__(self, toprule):

        if not isinstance(toprule, Rule):
            toprule = Rule(toprule)

        self.toprule = toprule


    def parse(self, input):
        """ Parse the given input and return the result of the parsing.

            An Exception will be raised if the parsing does not use the
            integrality of the input.
        """

        input = TextInput(input)
        results = Results()
        self.toprule.parse(input, results, None)

        if input.has_next():
            raise Exception(u("Finished parsing, but all the input was not consumed by the parser. Leftovers: '{0}'").format(input.input[input.pos:]))

        # Everything went fine, sending the results.
        if len(results) == 1:
            return results[0]
        return results


    def partial_parse(self, input, *args, **kwargs):
        """ Parse the given input and return only the result of the parsing,
            which is a tuple containing the (number of consumed characters, result of parsing).

            It can be used on some inputs on which we know the grammar can't or won't consume
            the totality of the input.
        """

        return self.toprule.parse(input)




import re
from .pwpeg import *

from functools import wraps
from itertools import chain


def _AllBut(but, escape):
    """
    """

    return OneOrMore(
        Either(
            Rule(escape, but).set_action(lambda escape, but: but),
            Rule(Not(but), Any())
        )
    ).set_skip(None)

AllBut = FunctionRule(_AllBut)
AllBut.set_name("All But")


def _Balanced(start, end, escape):
    """
    """

    balanced_inside = FunctionRule()

    def __balanced_inside():
        return Either(
            # Recurse on a balanced expression
            Rule(start, Optional(balanced_inside), end).set_action(lambda s, m, e: [s] + (m or []) + [e]),

            # Or simply gobble up characters that neither start nor end or their
            # backslashed version.
            AllBut.instanciate(Either(start, end), escape)
        )

    balanced_inside.set_fn(__balanced_inside)

    return Rule(start, ZeroOrMore(balanced_inside), end).set_action(lambda s, l, e: [s] + list(chain(*l)) + [e]).set_skip(None)

Balanced = FunctionRule(_Balanced)
Balanced.set_name("Balanced")


def _DelimitedBy(char, escape):
    """
    """

    return char, AllBut.instanciate(char, escape), char
DelimitedBy = FunctionRule(_DelimitedBy).set_skip(None)

def _RepeatingSeparated(rule, separator, at_least, at_most):
    """
    """
    def _repeat_action(first, rest):
        rest.insert(0, first)
        return rest

    if at_most == -1:
        # This is just so that it is equal to -1 in the end.
        at_most = 0

    if at_least == 0:
        return Optional(Rule(
            rule,
            Repetition(0, at_most - 1,
                separator,
                rule
            ).set_action(lambda l: map(lambda e: e[1], l))
        ).set_action(_repeat_action))

    return Rule(
        rule,
        Repetition(at_least - 1, at_most - 1,
            separator,
            rule
        ).set_action(lambda res: map(lambda x: x[1], res))
    ).set_action(_repeat_action)

RepeatingSeparated = FunctionRule(_RepeatingSeparated)

def _ZeroOrMoreSeparated(rules, sep):
    return RepeatingSeparated.instanciate(rules, sep, 0, -1)
ZeroOrMoreSeparated = FunctionRule(_ZeroOrMoreSeparated)


def _OneOrMoreSeparated(rules, sep):
    return RepeatingSeparated.instanciate(rules, sep, 1, -1)
OneOrMoreSeparated = FunctionRule(_OneOrMoreSeparated)


def _ExactlySeparated(how_much, rules, sep):
    return RepeatingSeparated.instanciate(rules, sep, how_much, how_much)
ExactlySeparated = FunctionRule(_ExactlySeparated)


def _RepetitionSeparated(at_least, at_most, rules, sep):
    return RepeatingSeparated.instanciate(rules, sep, at_least, at_most)
RepetitionSeparated = FunctionRule(_RepetitionSeparated)


re_more_indent = re.compile("[ \t]+")
re_lead = re.compile("^[ \t]+")
to_eol = Rule(re.compile("[ \t]*"))
empty_lines = Rule(re.compile("([ \t]*\n)*", re.M))


def la_tree_from_list_leftmost(leftmost, lst, action, idx=-1):
    return action(
        la_tree_from_list_leftmost(leftmost, lst, action, idx - 1),
        lst[idx][0],
        lst[idx][1]) if lst and idx >= -len(lst) else action(leftmost)

def ra_tree_from_list_rightmost(rightmost, lst, action, idx=0):
    """
    """

    return action(lst[idx][0],
                  lst[idx][1],
                  ra_tree_from_list_rightmost(rightmost, lst, action, idx + 1)) if lst and idx < len(lst) else action(rightmost, None, None)


def _LeftAssociative(op_FunctionRule, sub_FunctionRule):

    def reverse(leftmost, lst, idx=-1):
        """ Recursive function to make what we parsed as a left-associative succession of operators.
        """
        return la_tree_from_list_leftmost(leftmost, lst, action, idx)

    return Rule(sub_FunctionRule, ZeroOrMore(op_FunctionRule, sub_FunctionRule)).set_action(reverse)
LeftAssociative = FunctionRule(_LeftAssociative)

def _RightAssociative(op_FunctionRule, sub_FunctionRule):

    def _act(left, opt, idx=0):
        """ Recursive function to make what we parsed as a left-associative succession of operators.
        """

        return action(left) if not opt else action(left, opt[0], opt[1])

    self_recurse = R("RightAssociative", op_FunctionRule, sub_FunctionRule, action)

    return Rule(sub_FunctionRule, Optional(op_FunctionRule, self_recurse)).set_action(_act)
RightAssociative = FunctionRule(_RightAssociative)

####################################################################
#       Regexp Helpers

def delimitedby_regexp(char, escape):
    return re.compile(u("{0}({1}{0}|(?!{0}).)*{0}").format(re.escape(char), re.escape(escape)), re.DOTALL)

def allbut_regexp(patterns, escape):
    if not isinstance(patterns, list):
        patterns = [patterns]
    return re.compile("({1}({0})|(?!({0})).)+".format("|".join([re.escape(s) for s in patterns]), re.escape(escape)), re.DOTALL)


#!/usr/bin/env python
"""
    @author Christophe Eymard <christophe.eymard@ravelsoft.com>
"""

import re

from .pwpeg import *
from .helpers import *
from .pwast import *

re_regexp = re.compile("/(\\/|[^/])+/([idsmlux]+)?")

def _regexp_action(contents, flags):
    args = []

    args.append('\'' + contents.replace('\'', '\\\'') + '\'')

    if flags:
        fs = []
        for f in flags:
            fs.append("re.{0}".format(f.upper()))
        fs = " & ".join(fs)
        args.append(fs)

    return "re.compile({0})".format(", ".join(args))

def replace_regexps(t):
    return re_regexp.sub(lambda m: _regexp_action(m.group(0)[1:-1], m.group(2)), t)

def concat(arr):
    if arr is None or len(arr) == 0:
        return ""

    res = []
    for a in arr:
        if isinstance(a, list):
            res.append(concat(a))
        else:
            res.append(a)

    return "".join(res)

PIPE = "|"
SPACE = Rule(re.compile("\s*")).set_name("Whitespaces")
LINE_SPACE = re.compile("[ \t]*")
LPAREN = "("
RPAREN = ")"
LBRACKET = "["
RBRACKET = "]"
LBRACE = "{"
RBRACE = "}"
ARROW = "->"
DOLLAR = "$"
AMPS = "&"
EXCL = "!"
EQUAL = "="
COLON = ":"
COMMA = ","
STAR = "*"
QUESTION = "?"
PLUS = "+"
EOL = "\n"

############################

space_and_comments = re.compile("(\s+|#.*$)*", re.M)

identifier = Rule(re.compile("[a-zA-Z_][a-zA-Z0-9_]*")).set_name("Identifier")

to_eol = Rule(re.compile("[^\n]*", re.M))

empty_lines = Rule(re.compile("([ \t]*\n)*", re.M))

number = Rule(re.compile("-?[0-9]+")).set_action(lambda n: int(n))

code = Rule("%%", re.compile("((?!%%).)*", re.DOTALL), "%%").set_action(lambda b, t, e: t)

##############################################################################################

balanced_paren      = Balanced.instanciate("(", ")", "\\")
balanced_braces     = Balanced.instanciate("{", "}", "\\")
balanced_brackets   = Balanced.instanciate("[", "]", "\\")

# Re-concatenate the string.
string = Either(
    delimitedby_regexp("'", '\\'),
    delimitedby_regexp('"', '\\'),
    # Backslash quoted string
    Rule(re.compile("\\\\[^ \t\n\[\]\|\)]+")).set_action(lambda s: "'" + s[1:].replace('\'', '\\\'').replace('\\', '\\\\') + "'").set_name("Backslashed String")
).set_skip(None)
string.set_name("String")


regexp = Rule(
    #DelimitedBy.instanciate('/', '\\').set_action(lambda delimiter, escape, result: "".join(result)),
    delimitedby_regexp('/', '\\'),
    Optional(re.compile('[idsmlux]+'))
).set_skip(None).set_action(lambda d, f: _regexp_action(d[1:-1], f))
regexp.set_name("Regexp")


anything_inline = Rule(
    re.compile("[^\n]*")
)
anything_inline.set_name("Anything Inline")

blanks = Rule(re.compile("[ \t]+")).set_name("Blanks")

###############################
# ->
#   expr1
#   return action
def _action_multi_line():
    return Rule(
        Optional(SPACE),
        ARROW,
        Optional(LINE_SPACE),
        EOL,
        ZeroOrMore(
            Either(
                Rule(MemoRule(blanks), anything_inline, EOL),
                Rule(blanks, EOL),
            )
        )
    ).set_name("*Multi Line Action").set_action(lambda sp, arrow, more_space, eol, code: "\n".join([t[1] for t in code]))

action_multi_line = FunctionRule(_action_multi_line).set_skip(None)
#action_multi_line = _action_multi_line()
action_multi_line.set_name("Multi Line Action")

###############################
# -> action
action_single_line = Rule(
    ARROW,
    to_eol
).set_action(lambda arrow, line: line.strip()).set_skip(None).set_name("Single Line Action")

action_single_line.set_skip(None)

###############################
# {}, ->, -> \n
action = Either(
    action_multi_line,
    action_single_line
).set_name("Action")


#######################################################

###############################
# &{ return True }
predicate = Rule(
    balanced_braces
).set_action(lambda p: AstPredicate(concat(p[1:-1])))
predicate.set_name("Predicate")

###############################
# rule_ident
# rule_ident(balanced_paren, args)
rule_identifier = Rule(
    identifier,
    Optional(balanced_paren)
).set_action(lambda i, b: AstRuleDeclaration(i).set_args(replace_regexps(concat(b))))
rule_identifier.set_name("Rule Identifier")


###############################
# identifier:
label = Rule(identifier, COLON).set_action(lambda name, _2: name).set_name("Production Label")

###############################
# *, +, ?, <2>, <2,>, <,2>
repetition = Either(
    Rule("*").set_action(lambda x: (0, -1)),
    Rule("+").set_action(lambda x: (1, -1)),
    Rule("?").set_action(lambda x: (0,  1)),
    Rule("<", number, ">").set_action(lambda l, n, r: (n, n)),
    Rule("<", Optional(number), ",", Optional(number), ">").set_action(lambda l, fr, c, to, r: (-1 if fr is None else fr, -1 if to is None else to) )
)
repetition.set_name("Repetition Modifier")

rule_declaration = Rule().set_name("Rule Declaration")
production_group = Rule().set_name("Production Group")

###############################
# [ rule1 | rule2 ]
production_group_choices = Rule(
    OneOrMoreSeparated.instanciate(production_group, PIPE),
).set_action(lambda productions: AstProductionChoices(productions))
production_group_choices.set_name("Production Group Choices")


production = Rule(
    Optional(label),
    # The following is to prevent the parser from eating up a rule name
    # that serves in a subsequent rule declaration.
    # ie:  myrule* myrule2+     next_rule = ...
    # we prevent eating that:   ^^^^^^^^^
    Not(rule_declaration),
    Either(
        Either(
            regexp,
            string,
        ).set_action(lambda r: AstProduction(r)),
        Rule(rule_identifier).set_action(lambda d: AstRuleCall(d)).set_name("Rule Call"),
        Rule(LBRACKET, production_group_choices, RBRACKET).set_action(lambda _1, alts, _2: alts).set_name("Rule Choices")
    ),
    Optional(repetition)
)
production.set_action(lambda label, rule, rep: rule.set_label(label).set_repetition(rep))
production.set_name("Production")


###############################
# !rule &rule
# Basically, this is a look-ahead
look_ahead = Rule(
    Either("!", "&"),
    production,
    Optional(repetition)
)
look_ahead.set_action(lambda symbol, prod, rep: AstLookAhead(prod.set_repetition(rep), symbol))
look_ahead.set_name("Look Ahead")


###############################
# A single production rule
production_group.set_productions(
    OneOrMore(Either(
        look_ahead,
        production,
        predicate
    )),
    Optional(action)
).set_action(lambda rules, a: AstProductionGroup(rules).set_action(a))


####################### Rule declaration ###########################

###############################
# rule_name =
# rule_name(args) =
rule_declaration.set_productions(
    rule_identifier,
    Optional(Rule("skip", production).set_action(lambda _, rule: rule)),
    EQUAL
).set_name("Rule Declaration")
rule_declaration.set_action(lambda decl, skip, equal: decl.set_skip(skip))

grammarrule = Rule(
    rule_declaration,
    production_group_choices
).set_action(lambda decl, rules: decl.set_productions(rules))
grammarrule.set_name("Grammar Rule")

toplevel = Rule(
    Optional(code),
    OneOrMore(grammarrule),
    Optional(code),
    Optional(blanks)
)

toplevel.set_action(lambda code, rules, endcode, blanks: AstFile(code.strip() if code else "", rules, endcode.strip() if endcode else ""))
toplevel.set_skip(space_and_comments)
toplevel.set_name("Top Level")



from itertools import chain

from visitor import Visitor, indent

class PythonVisitor(Visitor):

    def __init__(self):
        super(PythonVisitor, self).__init__()
        self.nbfn = 0

    def compile_function(self, code, ctx, fnpattern="fn"):
        args = ", ".join(ctx.labels)

        if code.count("\n") == 0 and not code.startswith("return"):
            return "lambda {0}: ({1}) ".format(args, code)

        fname = "{0}_{1}".format(fnpattern, self.nbfn)
        self.nbfn += 1
        fbody = "def {0}({1}):\n{2}".format(fname, args, indent(code))

        ctx.functions.append(fbody)
        return fname

    def compile(self, node):
        self.visit(node)

        res = [
            "#!/usr/bin/env python",
            ""
            "from pwpeg import *",
            "from pwpeg.helpers import *",
            ""
        ]

        if self.code_start:
            res.append("\n##############################################################\n# Start of included code")
            res.append(self.code_start)
            res.append("\n# End of included code\n##############################################################")


        has_skips = False

        res.append("\n# Forward declaration of Regular rules")
        for name, sr in sorted(self.rules_simple.items()):
            res.append("{0} = Rule().set_name(\"{1}\")".format(name, sr.name))
            if sr.skip: has_skips = True
        res.append("")

        res.append("\n# Forward declaration of Function rules")
        for name, fr in sorted(self.rules_function.items()):
            res.append("{0} = FunctionRule().set_name(\"{1}\")".format(name, fr.name))
            if fr.skip: has_skips = True
        res.append("")

        if has_skips:
            res.append("\n# Skips")
            for name, r in sorted(chain(self.rules_simple.items(), self.rules_function.items())):
                if r.skip:
                    res.append("{0}.set_skip({1})".format(name, r.skip.name))
            res.append("")

        res.append("\n# Function Rules implementation")
        for name, fr in sorted(self.rules_function.items()):
            res.append("def _{0}{1}:".format(name, fr.args))
            for fn in fr.ctx.functions:
                res.append(indent(fn))
                res.append("")

            res.append("    return " + indent(fr.code)[4:])
            res.append("{0}.set_fn(_{0})\n".format(name))
        res.append("")

        res.append("\n# Simple Rules implementation")
        for name, sr in sorted(self.rules_simple.items()):
            for fn in sr.ctx.functions:
                res.append(fn)
                res.append("")

            if sr.code.startswith("Rule("):
                res.append("{0}.set_productions(".format(name) + sr.code[5:] + "\n")
            else:
                res.append("{0}.set_productions(".format(name) + sr.code + ")\n")
        res.append("")


        if self.code_end:
            res.append("\n##############################################################\n# Start of included code")
            res.append(self.code_end)
            res.append("\n# End of included code\n##############################################################")

        return "\n".join(res)




#!/usr/bin/env python

from distutils.core import setup

setup(name='PwPEG',
      version='0.3.1',
      description='Python PEG Parsing Framework',
      author='Christophe Eymard',
      author_email='christophe.eymard@ravelsoft.com',
      url='http://pwpeg.github.com/',
      packages=['pwpeg'],
      scripts=['scripts/pwpeg']
     )



def indent(txt):
    return "\n".join(["    " + t for t in txt.split("\n")])

class Context(object):
    def __init__(self):
        self.labels = []
        self.predicates = []
        self.action = None
        self.functions = []

    def add_rule(self, rule):
        if rule.label:
            self.labels.append(rule.label)
        else:
            self.labels.append("_{0}".format(len(self.labels)))

    def merge(self, ctx):
        self.functions += ctx.functions


class Visitor(object):
    def __init__(self):
        self.rules = dict()
        self.rules_simple = dict()
        self.rules_function = dict()
        self.code_start = ""
        self.code_end = ""
        self.tokens = set()

    def visit(self, node, *args, **kwargs):
        return node.accept(self, *args, **kwargs)

    #######################################################################

    def visit_Repetition(self, node):
        if node.repetition:
            _from, _to = node.repetition
            repeat = "Repetition({0}, {1}, {{0}})".format(_from, _to)

            if _from == 0 and _to == -1:
                repeat = "ZeroOrMore({0})"
            if _from == 1 and _to == -1:
                repeat = "OneOrMore({0})"
            if _from == 0 and _to == 1:
                repeat = "Optional({0})"

            return repeat.format(node.code)
        else:
            return node.code

    def visit_AstLookAhead(self, node, ctx):
        nctx = Context()
        return ("Not(" if node.symbol == "!" else "And(") + self.visit(node.production, nctx) + ")"

    def visit_AstPredicate(self, node, ctx):
        return self.compile_function(node.code, ctx)

    def visit_AstRuleCall(self, node, ctx):
        ctx.add_rule(node)
        node.code = node.decl.name + (".instanciate" + node.decl.args if node.decl.args else "")
        return self.visit_Repetition(node)

    def visit_AstProduction(self, node, ctx):
        ctx.add_rule(node)
        return self.visit_Repetition(node)

    def visit_AstProductionGroup(self, node, ctx):
        node.labels = []

        c = Context()
        subnodes = [self.visit(p, c) for p in node.rules]
        if len(subnodes) > 1:
            node.code = "Rule(\n" + ",\n".join(map(indent, subnodes)) + "\n)"
        else:
            node.code = subnodes[0]

        if node.action:
            node.code += ".set_action({0})".format(self.compile_function(node.action, c, "action"))

        ctx.merge(c)

        return self.visit_Repetition(node)

    def visit_AstProductionChoices(self, node, ctx):
        res = []

        ctx.add_rule(node)

        c = Context()

        if len(node.rules) > 1:
            res.append("Either(")
            res.append(",\n".join(map(indent, [self.visit(p, c) for p in node.rules])))
            res.append(")")
        else:
            res.append(self.visit(node.rules[0], c))

        node.code = "\n".join(res)

        if node.action:
            node.code += ".set_action({0})".format(self.compile_function(node.action, c, "action"))

        ctx.merge(c)
        return self.visit_Repetition(node)

    def visit_AstRuleDeclaration(self, node):
        if node.name in self.rules:
            raise Exception("Can't redefine existing rule {0}".format(node.name))

        if node.args:
            self.rules_function[node.name] = node
        else:
            self.rules_simple[node.name] = node

        self.rules[node.name] = node
        node.ctx = Context()
        node.code = self.visit(node.productions, node.ctx)

    def visit_AstFile(self, node):
        for r in node.rules:
            self.visit(r)
        self.code_start = node.code_start
        self.code_end = node.code_end
